#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 19:33:20 2018

@author: yoelr
"""
import numpy as np
from thermosteam import MultiStream, Stream
from .design_tools.specification_factors import  (
    distillation_column_material_factors,
    tray_material_factor_functions,
    distillation_tray_type_factor,
    material_densities_lb_per_in3)
from .design_tools.column_design import (
    compute_tower_height,
    compute_tower_diameter,
    compute_tower_weight,
    compute_tower_wall_thickness,
    compute_flow_parameter,
    compute_max_vapor_velocity,
    compute_max_capacity_parameter,
    compute_downcomer_area_fraction,
    compute_murphree_stage_efficiency,
    compute_purchase_cost_of_trays,
    compute_purchase_cost_of_tower)
from .. import Unit
from scipy.optimize import brentq
from ._hx import HXutility
import matplotlib.pyplot as plt

def compute_stages_McCabeThiele(P, operating_line,
                                x_stages, y_stages, T_stages,
                                x_limit, solve_Ty):
    """
    Find the specifications at every stage of the of the operating line before
    the maximum liquid molar fraction. Append the light key liquid molar
    fraction, light key vapor molar fraction, and stage temperatures to
    x_stages, y_stages and T_stages respectively.
    
    Parameters
    ----------
    P : float
        Pressure [Pa].
    operating_line : function
                     Should return the liquid molar fraction of the light
                     key given its vapor molar fraction.
    x_stages : list
               Liquid molar compositions at each stage. Last element
               should be the starting point for the next stage.
    y_stages : list
               Vapor molar compositions at each stage. Last element 
               should be the starting point for the next stage.
    x_limit : float
              Maximum value of liquid composition before algorithm stops.
    T_stages : list
               Temperature at each stage.
    solve_Ty : function
               Should return T and y given x.
        
    """
    i = 0
    yi = y_stages[-1]
    xi = x_stages[-1]
    while xi < x_limit:
        if i > 100:
            raise RuntimeError('cannot meet specifications! stages > 100')
        i += 1
        # Go Up
        x = np.array((xi, 1-xi))
        T, y = solve_Ty(x, P)
        yi = y[0]
        y_stages.append(yi)
        T_stages.append(T)
        # Go Right
        xi = operating_line(yi)
        if xi > x_limit:
            xi = x_limit
        x_stages.append(xi)

# %% Distillation

class Distillation(Unit):
    r"""
    Create a Distillation column that assumes all light and heavy non keys
    separate to the top and bottoms product respectively. McCabe-Thiele
    analysis is used to find both the number of stages and the reflux ratio
    given a ratio of actual reflux to minimum reflux [1]_. This assumption
    is good for both binary distillation of highly polar compounds and
    ternary distillation assuming complete separation of light non-keys
    and heavy non-keys with large differences in boiling points. Preliminary
    analysis showed that the theoretical number of stages using this method
    on Methanol/Glycerol/Water systems is off by less than +-1 stage. Other
    methods, such as the Fenske-Underwood-Gilliland method, are more suitable
    for hydrocarbons. The Murphree efficiency is based on the modified
    O'Connell correlation [2]_. The diameter is based on tray separation
    and flooding velocity [1]_. Purchase costs are based on correlations
    by Mulet et al. [3]_ [4]_ as compiled by Warren et. al. [5]_.

    Parameters
    ----------
    ins : streams
        Inlet fluids to be mixed into the feed stage.
    outs : stream sequence
        * [0] Distillate
        * [1] Bottoms product
    LHK : tuple[str]
        Light and heavy keys.
    y_top : float
        Molar fraction of light key in the distillate.
    x_bot : float
        Molar fraction of light key in the bottoms.
    k : float
        Ratio of reflux to minimum reflux.
    P=101325 : float
        Operating pressure [Pa].
    vessel_material : str, optional
        Vessel construction material. Defaults to 'Carbon steel'.
    tray_material : str, optional
        Tray construction material. Defaults to 'Carbon steel'.
    tray_type='Sieve' : 'Sieve', 'Valve', or 'Bubble cap'
        Tray type.
    tray_spacing=450 : float
        Typically between 152 to 915 mm.
    stage_efficiency=None : 
        User enforced stage efficiency. If None, stage efficiency is
        calculated by the O'Connell correlation [2]_.
    velocity_fraction=0.8 : float
        Fraction of actual velocity to maximum velocity allowable before flooding.
    foaming_factor=1.0 : float
        Must be between 0 to 1.
    open_tray_area_fraction=0.1 : float
        Fraction of open area to active area of a tray.
    downcomer_area_fraction=None : float
        Enforced fraction of downcomer area to net (total) area of a tray.
        If None, estimate ratio based on Oliver's estimation [1]_.
    is_divided=False : bool
        True if the stripper and rectifier are two separate columns.

    References
    ----------
    .. [1] J.D. Seader, E.J. Henley, D.K. Roper. (2011)
        Separation Process Principles 3rd Edition. John Wiley & Sons, Inc. 

    .. [2] M. Duss, R. Taylor. (2018)
        Predict Distillation Tray Efficiency. AICHE 
    
    .. [3] Mulet, A., A. B. Corripio, and L. B. Evans. (1981a).
        Estimate Costs of Pressure Vessels via Correlations.
        Chem. Eng., 88(20), 145–150.

    .. [4] Mulet, A., A.B. Corripio, and L.B.Evans. (1981b).
        Estimate Costs of Distillation and Absorption Towers via Correlations.
        Chem. Eng., 88(26), 77–82.

    .. [5] Seider, W. D., Lewin,  D. R., Seader, J. D., Widagdo, S., Gani, R.,
        & Ng, M. K. (2017). Product and Process Design Principles. Wiley.
        Cost Accounting and Capital Cost Estimation (Chapter 16)    

    Examples
    --------
    Binary distillation assuming 100% separation on non-keys:
    
    >>> from biosteam.units import Distillation
    >>> from thermosteam import Stream, Chemicals, settings
    >>> chemicals = Chemicals(['Water', 'Methanol', 'Glycerol'])
    >>> settings.set_thermo(chemicals)
    >>> feed = Stream('feed', flow=(80, 100, 25))
    >>> bp = feed.bubble_point_at_P()
    >>> feed.T = bp.T # Feed at bubble point T
    >>> D1 = Distillation('D1', ins=feed,
    ...                   outs=('distillate', 'bottoms_product'),
    ...                   LHK=('Methanol', 'Water'),
    ...                   y_top=0.99, x_bot=0.01, k=2)
    >>> D1.is_divided = True
    >>> D1.simulate()
    >>> # See all results
    >>> D1.show(T='degC', P='atm', composition=True)
    Distillation: D1
    ins...
    [0] feed
        phase: 'l', T: 76.129 degC, P: 1 atm
        composition: Water     0.39
                     Methanol  0.488
                     Glycerol  0.122
                     --------  205 kmol/hr
    outs...
    [0] distillate
        phase: 'g', T: 64.91 degC, P: 1 atm
        composition: Water     0.01
                     Methanol  0.99
                     --------  100 kmol/hr
    [1] bottoms_product
        phase: 'l', T: 100.06 degC, P: 1 atm
        composition: Water     0.754
                     Methanol  0.00761
                     Glycerol  0.239
                     --------  105 kmol/hr
    >>> D1.results()
    Distillation                                    Units        D1
    Cooling water       Duty                        kJ/hr -5.11e+06
                        Flow                      kmol/hr  3.49e+03
                        Cost                       USD/hr      1.71
    Low pressure steam  Duty                        kJ/hr  9.49e+06
                        Flow                      kmol/hr       244
                        Cost                       USD/hr      58.1
    Design              Theoretical feed stage                    9
                        Theoretical stages                       13
                        Minimum reflux                        0.687
                        Reflux                                 1.37
                        Rectifier stages                         15
                        Stripper stages                          13
                        Rectifier height                       34.7
                        Stripper height                        31.7
                        Rectifier diameter                     3.93
                        Stripper diameter                      3.36
                        Rectifier wall thickness               0.25
                        Stripper wall thickness                0.25
                        Rectifier weight                   4.79e+03
                        Stripper weight                    3.74e+03
    Purchase cost       Rectifier trays               USD   1.5e+04
                        Stripper trays                USD  1.28e+04
                        Rectifier tower               USD  7.62e+04
                        Stripper tower                USD  6.53e+04
                        Condenser                     USD  3.41e+04
                        Boiler                        USD  2.64e+04
    Total purchase cost                               USD   2.3e+05
    Utility cost                                   USD/hr      59.8
    
    """
    line = 'Distillation'
    _N_heat_utilities = 0
    units = {'Minimum reflux': 'Ratio',
             'Reflux': 'Ratio',
             'Rectifier height': 'ft',
             'Rectifier diameter': 'ft',
             'Rectifier wall thickness': 'in',
             'Rectifier weight': 'lb',
             'Stripper height': 'ft',
             'Stripper diameter': 'ft',
             'Stripper wall thickness': 'in',
             'Stripper weight': 'lb',
             'Height': 'ft',
             'Diameter': 'ft',
             'Wall thickness': 'in',
             'Weight': 'lb'}
    # Bare module factor
    BM = 4.3
    
    # [dict] Bounds for results
    _bounds = {'Diameter': (3., 24.),
               'Height': (27., 170.),
               'Weight': (9000., 2.5e6)}
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None,
                P=101325, *, LHK, y_top, x_bot, k,
                vessel_material='Carbon steel',
                tray_material='Carbon steel',
                tray_type='Sieve',
                tray_spacing=450,
                stage_efficiency=None,
                velocity_fraction=0.8,
                foaming_factor=1.0,
                open_tray_area_fraction=0.1,
                downcomer_area_fraction=None,
                is_divided=False):
        Unit.__init__(self, ID, ins, outs, thermo)
        
        # Operation specifications
        self.P = P
        self.LHK = LHK
        self.y_top = y_top
        self.x_bot = x_bot
        self.k = k
        
        # Construction specifications
        self.vessel_material = vessel_material
        self.tray_type = tray_type
        self.tray_material = tray_material
        self.tray_spacing = tray_spacing
        self.stage_efficiency = stage_efficiency
        self.velocity_fraction = velocity_fraction
        self.foaming_factor = foaming_factor
        self.open_tray_area_fraction = open_tray_area_fraction
        self.downcomer_area_fraction = downcomer_area_fraction
        self.is_divided = is_divided
        
        # Setup components
        thermo = self.thermo
        self.condenser = HXutility(None,
                                    ins=Stream(None, phase='g', thermo=thermo),
                                    outs=MultiStream(None, thermo=thermo),
                                    thermo=thermo)
        self.boiler = HXutility(None,
                                 ins=Stream(None, thermo=thermo),
                                 outs=MultiStream(None, thermo=thermo),
                                 thermo=thermo)
        self.heat_utilities = self.condenser.heat_utilities + self.boiler.heat_utilities
        self.condensate = Stream(None, thermo=thermo)
        self.boil_up = Stream(None, phase='g', thermo=thermo)
        self.precondensate =  Stream(None)
        self._McCabeThiele_args = np.zeros(6)
    
    @property
    def LHK(self):
        """tuple[str, str] Light and heavy keys."""
        return self._LHK
    @LHK.setter
    def LHK(self, LHK):
        # Set light non-key and heavy non-key indices
        self._LHK = LHK = tuple(LHK)
        chemicals = self.chemicals
        LHK_chemicals = LK_chemical, HK_chemical = self.chemicals.retrieve(LHK)
        Tb_light = LK_chemical.Tb
        Tb_heavy = HK_chemical.Tb
        LNK = []
        HNK = []
        if Tb_light > Tb_heavy:
            raise ValueError(f"light key must be lighter than heavy key")
        for chemical in chemicals:
            Tb = chemical.Tb
            if not Tb:
                HNK.append(chemical.ID)
            elif Tb < Tb_light:
                LNK.append(chemical.ID)
            elif Tb > Tb_heavy:
                HNK.append(chemical.ID)
            elif chemical not in LHK_chemicals:
                raise ValueError(f"intermediate volatile specie, '{chemical}', between light and heavy key, ['{LK_chemical}', '{HK_chemical}']")
        self._LNK = tuple(LNK)
        self._HNK = tuple(HNK)
    
    @property
    def y_top(self):
        """Light key composition of at the distillate."""
        return self._y_top
    @y_top.setter
    def y_top(self, y_top):
        self._y_top = y_top
        self._y = np.array([y_top, 1-y_top])
    
    @property
    def x_bot(self):
        """Light key composition at the bottoms product."""
        return self._x_bot
    @x_bot.setter
    def x_bot(self, x_bot):
        self._x_bot = x_bot
        self._x = np.array([x_bot, 1-x_bot])
    
    @property
    def tray_spacing(self):
        return self._TS
    @tray_spacing.setter
    def tray_spacing(self, TS):
        """Tray spacing (225-600 mm)."""
        self._TS = TS
    
    @property
    def stage_efficiency(self):
        """Enforced user defined stage efficiency."""
        return self._E_eff
    @stage_efficiency.setter
    def stage_efficiency(self, E_eff):
        self._E_eff = E_eff
    
    @property
    def velocity_fraction(self):
        """Fraction of actual velocity to maximum velocity allowable before flooding."""
        return self._f
    @velocity_fraction.setter
    def velocity_fraction(self, f):
        self._f = f
    
    @property
    def foaming_factor(self):
        """Foaming factor (0 to 1)."""
        return self._F_F
    @foaming_factor.setter
    def foaming_factor(self, F_F):
        if not 0 <= F_F <= 1:
            raise ValueError(f"foaming_factor must be between 0 and 1, ({F_F} given).")
        self._F_F = F_F
    
    @property
    def open_tray_area_fraction(self):
        """Fraction of open area, A_h, to active area, A_a."""
        return self._A_ha
    @open_tray_area_fraction.setter
    def open_tray_area_fraction(self, A_ha):
        self._A_ha = A_ha
    
    @property
    def downcomer_area_fraction(self):
        """Enforced fraction of downcomer area to net (total) area.
        If None, the fraction is estimated based on heuristics."""
        return self._A_dn
    @downcomer_area_fraction.setter
    def downcomer_area_fraction(self, A_dn):
        self._A_dn = A_dn
    
    @property
    def tray_type(self):
        """Default 'Sieve'"""
        return self._tray_type
    @tray_type.setter
    def tray_type(self, tray_type):
        if tray_type in distillation_tray_type_factor:
            self._tray_type = tray_type
            self._F_TT = distillation_tray_type_factor[tray_type]
        else:
            raise ValueError("tray type must be one of the following: "
                            f"{', '.join(distillation_tray_type_factor)}")
        
    @property
    def tray_material(self):
        """Default 'Carbon steel'"""
        return self._tray_material
    @tray_material.setter
    def tray_material(self, tray_material):
        if tray_material in tray_material_factor_functions:
            self._tray_material = tray_material
            self._F_TM_function = tray_material_factor_functions[tray_material]
        else:
            raise ValueError("tray material must be one of the following: "
                            f"{', '.join(tray_material_factor_functions)}")
        

    @property
    def vessel_material(self):
        """Default 'Carbon steel'"""
        return self._vessel_material
    @vessel_material.setter
    def vessel_material(self, vessel_material):
        if vessel_material in distillation_column_material_factors:
            self._vessel_material = vessel_material
            self._F_VM = distillation_column_material_factors[vessel_material]
        else:
            raise ValueError("vessel material must be one of the following: "
                            f"{', '.join(distillation_column_material_factors)}")
    
    @property
    def is_divided(self):
        """[bool] True if the stripper and rectifier are two separate columns."""
        return self._is_divided
    @is_divided.setter
    def is_divided(self, is_divided):
        self._is_divided = is_divided
    
    def _run_mass_balance(self):
        # Get all important flow rates (both light and heavy keys and non-keys)
        get_index = self.chemicals.get_index
        LHK_index = get_index(self._LHK)
        LNK_index = get_index(self._LNK)
        HNK_index = get_index(self._HNK)
        mol = self.mol_in
        LHK_mol = mol[LHK_index]
        HNK_mol = mol[HNK_index]
        LNK_mol = mol[LNK_index]

        # Set light and heavy keys by lever rule
        light, heavy = LHK_mol
        LHK_F_mol = light + heavy
        zf = light/LHK_F_mol
        split_frac = (zf-self.x_bot)/(self.y_top-self.x_bot)
        top_net = LHK_F_mol*split_frac

        # Set output streams
        vap, liq = self.outs
        vap.mol[LHK_index] = vap_LHK_mol = top_net * self._y
        liq.mol[LHK_index] = LHK_mol - vap_LHK_mol
        vap.mol[LNK_index] = LNK_mol
        liq.mol[HNK_index] = HNK_mol
    
    def _run(self):
        self._run_mass_balance()
        vap, liq = self.outs
        vap.phase = 'g'
        liq.phase = 'l'
        vap.P = liq.P = self.P
        self._condensate_dew_point = dp = vap.dew_point_at_P()
        self._boilup_bubble_point = bp = liq.bubble_point_at_P()
        liq.T = bp.T
        vap.T = dp.T

    def _simulate_condenser(self):
        distillate = self.outs[0]
        condensate = self.condensate # Abstract attribute
        condenser = self.condenser
        s_in = condenser.ins[0]
        s_in.mol[:] = distillate.mol + condensate.mol
        s_in.T = distillate.T
        s_in.P = distillate.P
        ms1 = condenser.outs[0]
        ms1.imol['l'] = condensate.mol
        ms1.T = condensate.T
        ms1.P = condensate.P
        ms1.imol['g'] = distillate.mol
        condenser._design()
        condenser._cost()
        
    def _simulate_boiler(self):
        bottoms = self.outs[1]
        boil_up = self.boil_up # Abstract attribute
        boiler = self.boiler
        s_in = boiler.ins[0]
        s_in.copy_like(bottoms)
        s_in.mol += boil_up.mol
        ms1 = boiler.outs[0]
        ms1.T = boil_up.T
        ms1.P = boil_up.P
        ms1.imol['g'] = boil_up.mol
        ms1.imol['l'] = bottoms.mol
        if hasattr(self, 'condenser'):
            boiler._design(self.H_out - self.H_in - self.condenser.Q)
            boiler._cost()
        else:
            boiler._design()
            boiler._cost()  
    
    def _simulate_components(self): 
        # Cost condenser
        self._simulate_condenser()
        self.purchase_costs['Condenser'] = self.condenser.purchase_costs['Heat exchanger']
        
        # Cost boiler
        self._simulate_boiler()
        self.purchase_costs['Boiler'] = self.boiler.purchase_costs['Heat exchanger']
    
    def compute_N_stages(self):
        """Return a tuple with the actual number of stages for the rectifier and the stripper."""
        vap, liq = self.outs
        Design = self.design_results
        x_stages = self._x_stages
        y_stages = self._y_stages
        R = Design['Reflux']
        N_stages = Design['Theoretical stages']
        feed_stage = Design['Theoretical feed stage']
        liq_mol = self._feed_liqmol
        vap_mol = self._feed_vapmol
        
        E_eff = self.stage_efficiency
        if E_eff:
            E_rectifier = E_stripper = E_eff
        else:    
            # Calculate Murphree Efficiency for rectifying section
            vap_F_mol = vap.F_mol
            dp = self._condensate_dew_point
            condensate_x_mol = dp.x
            self._L_Rmol = L_Rmol = R*vap_F_mol
            self._V_Rmol = V_Rmol = (R+1)*vap_F_mol
            condensate = self.condensate
            condensate.imol[dp.IDs] = condensate_x_mol * L_Rmol
            condensate.T = vap.T
            condensate.P = vap.P
            mu = condensate.get_property('mu', 'mPa*s')
            K_light = y_stages[-1]/x_stages[-1] 
            K_heavy = (1-y_stages[-1])/(1-x_stages[-1])
            alpha = K_light/K_heavy
            E_rectifier = compute_murphree_stage_efficiency(mu, alpha, L_Rmol, V_Rmol)
            
            # Calculate Murphree Efficiency for stripping section
            mu = liq.get_property('mu', 'mPa*s')
            self._V_Smol = V_Smol = (R+1)*vap_F_mol - sum(vap_mol)
            self._L_Smol = L_Smol = R*vap_F_mol + sum(liq_mol) 
            K_light = y_stages[0]/x_stages[0] 
            K_heavy = (1-y_stages[0])/(1-x_stages[0] )
            alpha = K_light/K_heavy
            E_stripper = compute_murphree_stage_efficiency(mu, alpha, L_Smol, V_Smol)
            
        # Calculate actual number of stages
        mid_stage = feed_stage - 0.5
        N_rectifier = np.ceil(mid_stage/E_rectifier)
        N_stripper = np.ceil((N_stages-mid_stage)/E_stripper)
        return N_rectifier, N_stripper

    def _run_McCabeThiele(self):
        distillate, bottoms = self.outs
        chemicals = self.chemicals
        LHK = self._LHK
        LHK_index = chemicals.get_index(LHK)

        # Feed light key mol fraction
        liq_mol = np.zeros(chemicals.size)
        vap_mol = liq_mol.copy()
        for s in self.ins:
            if isinstance(s, MultiStream):
                liq_mol += s.imol['l']
                vap_mol += s.imol['g']
            elif s.phase == 'g':
                vap_mol += s.mol
            elif s.phase.lower() == 'l':
                liq_mol += s.mol
            elif s.phase == 's': pass
            else:
                raise RuntimeError(f'invalid phase encountered in {repr(s)}')
        self._feed_liqmol = liq_mol
        self._feed_vapmol = vap_mol
        LHK_mol = liq_mol[LHK_index] + vap_mol[LHK_index]
        LHK_F_mol = LHK_mol.sum()
        zf = LHK_mol[0]/LHK_F_mol
        
        # Get feed quality
        q = liq_mol[LHK_index].sum()/LHK_F_mol
        
        # Main arguments
        P = self.P
        k = self.k
        y_top = self.y_top
        x_bot = self.x_bot
        
        # Cache
        old_args = self._McCabeThiele_args
        args = np.array([P, k, y_top, x_bot, q, zf])
        if (abs(old_args - args) < np.array([50, 1e-5, 1e-6, 1e-6, 1e-6, 1e-6], float)).all(): return
        self._McCabeThiele_args = args
        
        # Get R_min and the q_line 
        if q == 1:
            q = 1 - 1e-5
        q_line = lambda x: q*x/(q-1) - zf/(q-1)
        self._q_line_args = dict(q=q, zf=zf)
        
        solve_Ty = bottoms.get_bubble_point(LHK).solve_Ty
        Rmin_intersection = lambda x: q_line(x) - solve_Ty(np.array((x, 1-x)), P)[1][0]
        x_Rmin = brentq(Rmin_intersection, 0, 1)
        y_Rmin = q_line(x_Rmin)
        m = (y_Rmin-y_top)/(x_Rmin-y_top)
        Rmin = m/(1-m)
        if Rmin <= 0:
            R = 0.0001*k
        else:
            R = Rmin*k

        # Rectifying section: Inntersects q_line with slope given by R/(R+1)
        m1 = R/(R+1)
        b1 = y_top-m1*y_top
        rs = lambda y: (y - b1)/m1 # -> x
        
        # y_m is the solution to lambda y: y - q_line(rs(y))
        self._y_m = y_m = (q*b1 + m1*zf)/(q - m1*(q-1))
        self._x_m = x_m = rs(y_m)
        
        # Stripping section: Intersects Rectifying section and q_line and beggins at bottoms liquid composition
        m2 = (x_bot-y_m)/(x_bot-x_m)
        b2 = y_m-m2*x_m
        ss = lambda y: (y-b2)/m2 # -> x        
        
        # Data for staircase
        self._x_stages = x_stages = [x_bot]
        self._y_stages = y_stages = [x_bot]
        self._T_stages = T_stages = []
        compute_stages_McCabeThiele(P, ss, x_stages, y_stages, T_stages, x_m, solve_Ty)
        yi = y_stages[-1]
        xi = rs(yi)
        x_stages[-1] = xi if xi < 1 else 0.99999
        compute_stages_McCabeThiele(P, rs, x_stages, y_stages, T_stages, y_top, solve_Ty)
        
        # Find feed stage
        for i in range(len(y_stages)-1):
            j = i+1
            if y_stages[i] < y_m and y_stages[j] > y_m:
                feed_stage = i+1
        stages = len(x_stages)
        
        # Results
        Design = self.design_results
        Design['Theoretical feed stage'] = stages - feed_stage
        Design['Theoretical stages'] = stages
        Design['Minimum reflux'] = Rmin
        Design['Reflux'] = R
        
    def _design(self):
        self._run_McCabeThiele()
        distillate, bottoms = self._outs
        Design = self.design_results
        R = Design['Reflux']
        Rstages, Sstages = self.compute_N_stages()
        is_divided = self.is_divided
        TS = self._TS
        
        ### Get diameter of rectifying section based on top plate ###
        
        condensate = self.condensate
        rho_L = condensate.rho
        sigma = condensate.get_property('sigma', 'dyn/cm')
        L = condensate.F_mass
        V = L*(R+1)/R
        precondensate = self.precondensate
        precondensate.copy_like(distillate)
        precondensate.mol *= R+1
        V_vol = precondensate.get_total_flow('m^3/s')
        rho_V = distillate.rho
        F_LV = compute_flow_parameter(L, V, rho_V, rho_L)
        C_sbf = compute_max_capacity_parameter(TS, F_LV)
        F_F = self._F_F
        A_ha = self._A_ha
        U_f = compute_max_vapor_velocity(C_sbf, sigma, rho_L, rho_V, F_F, A_ha)
        A_dn = self._A_dn
        if A_dn is None:
           self._A_dn = A_dn = compute_downcomer_area_fraction(F_LV)
        f = self._f
        R_diameter = compute_tower_diameter(V_vol, U_f, f, A_dn)
        
        ### Get diameter of stripping section based on feed plate ###
        
        V_mol = self._V_Smol
        rho_L = bottoms.rho
        bp = self._boilup_bubble_point
        boil_up_flow = bp.y * V_mol
        boil_up = self.boil_up
        boil_up.T = bottoms.T
        boil_up.P = bottoms.P
        boil_up.imol[bp.IDs] = boil_up_flow
        V = boil_up.F_mass
        V_vol = boil_up.get_total_flow('m^3/s')
        rho_V = boil_up.rho
        L = bottoms.F_mass # To get liquid going down
        F_LV = compute_flow_parameter(L, V, rho_V, rho_L)
        C_sbf = compute_max_capacity_parameter(TS, F_LV)
        sigma = condensate.get_property('sigma', 'dyn/cm')
        U_f = compute_max_vapor_velocity(C_sbf, sigma, rho_L, rho_V, F_F, A_ha)
        A_dn = self._A_dn
        if A_dn is None:
            A_dn = compute_downcomer_area_fraction(F_LV)
        S_diameter = compute_tower_diameter(V_vol, U_f, f, A_dn)
        Po = self.P * 0.000145078 # to psi
        rho_M = material_densities_lb_per_in3[self.vessel_material]
        
        if is_divided:
            Design['Rectifier stages'] = Rstages
            Design['Stripper stages'] =  Sstages
            Design['Rectifier height'] = H_R = compute_tower_height(TS, Rstages-1)
            Design['Stripper height'] = H_S = compute_tower_height(TS, Sstages-1)
            Design['Rectifier diameter'] = R_diameter
            Design['Stripper diameter'] = S_diameter
            Design['Rectifier wall thickness'] = tv = compute_tower_wall_thickness(Po, R_diameter, H_R)
            Design['Stripper wall thickness'] = tv = compute_tower_wall_thickness(Po, S_diameter, H_S)
            Design['Rectifier weight'] = compute_tower_weight(R_diameter, H_R, tv, rho_M)
            Design['Stripper weight'] = compute_tower_weight(S_diameter, H_S, tv, rho_M)
        else:
            Design['Actual stages'] = Rstages + Sstages
            Design['Height'] = H = compute_tower_height(TS, Rstages+Sstages-2)
            Design['Diameter'] = Di = max((R_diameter, S_diameter))
            Design['Wall thickness'] = tv = compute_tower_wall_thickness(Po, Di, H)
            Design['Weight'] = compute_tower_weight(Di, H, tv, rho_M)
        
    def _cost(self):
        Design = self.design_results
        Cost = self.purchase_costs
        F_TT = self._F_TT
        F_VM = self._F_VM
        if self.is_divided:
            # Number of trays assuming a partial condenser
            N_RT = Design['Rectifier stages'] - 1
            Di_R = Design['Rectifier diameter']
            F_TM = self._F_TM_function(Di_R)
            Cost['Rectifier trays'] = compute_purchase_cost_of_trays(N_RT, Di_R, F_TT, F_TM)
            N_ST = Design['Stripper stages'] - 1
            Di_S = Design['Stripper diameter']
            F_TM = self._F_TM_function(Di_R)
            Cost['Stripper trays'] = compute_purchase_cost_of_trays(N_ST, Di_S, F_TT, F_TM)
            
            # Cost vessel assuming T < 800 F
            W_R = Design['Rectifier weight'] # in lb
            H_R = Design['Rectifier height']*3.28 # in ft
            Cost['Rectifier tower'] = compute_purchase_cost_of_tower(Di_R, H_R, W_R, F_VM)
            W_S = Design['Stripper weight'] # in lb
            H_S = Design['Stripper height']*3.28 # in ft
            Cost['Stripper tower'] = compute_purchase_cost_of_tower(Di_S, H_S, W_S, F_VM)
        else:
            # Cost trays assuming a partial condenser
            N_T = Design['Actual stages'] - 1
            Di = Design['Diameter']
            F_TM = self._F_TM_function(Di)
            Cost['Trays'] = compute_purchase_cost_of_trays(N_T, Di, F_TT, F_TM)
            
            # Cost vessel assuming T < 800 F
            W = Design['Weight'] # in lb
            L = Design['Height']*3.28 # in ft
            Cost['Tower'] = compute_purchase_cost_of_tower(Di, L, W, F_VM)
        self._simulate_components()
    
    def _plot_stages(self):
        """Plot stages, graphical aid line, and equilibrium curve. The plot does not include operating lines nor a legend."""
        vap, liq = self.outs
        if not hasattr(self, 'x_stages'):
            self._design()
        x_stages = self._x_stages
        y_stages = self._y_stages
        LHK = self.LHK
        LK = self.LHK[0]
        P = self.P
        
        # Equilibrium data
        x_eq = np.linspace(0, 1, 100)
        y_eq = np.zeros(100)
        T = np.zeros(100)
        n = 0
        
        bp = vap.get_bubble_point(IDs=LHK)
        solve_Ty = bp.solve_Ty
        for xi in x_eq:
            T[n], y = solve_Ty(np.array([xi, 1-xi]), P)
            y_eq[n] = y[0]
            n += 1
            
        # Set-up graph
        plt.figure()
        plt.xticks(np.arange(0, 1.1, 0.1), fontsize=12)
        plt.yticks(fontsize=12)
        plt.xlabel('x (' + LK + ')', fontsize=16)
        plt.ylabel('y (' + LK + ')', fontsize=16)
        plt.xlim([0, 1])
        
        # Plot stages
        x_stairs = []
        for x in x_stages:
            x_stairs.append(x)
            x_stairs.append(x)
            
        y_stairs = []
        for y in y_stages:
            y_stairs.append(y)
            y_stairs.append(y)
        x_stairs.pop(-1)
        x_stairs.insert(0, y_stairs[0])
        plt.plot(x_stairs, y_stairs, '--')
        
        # Graphical aid line
        plt.plot([0, 1], [0, 1])
        
        # Vapor equilibrium graph
        plt.plot(x_eq, y_eq, lw=2)
    
    def plot_stages(self):
        """Plot the McCabe Thiele Diagram."""
        # Plot stages, graphical aid and equilibrium curve
        self._plot_stages()
        vap, liq = self.outs
        Design = self.design_results
        if not hasattr(self, '_x_stages'): self._design()
        q_args = self._q_line_args
        zf = q_args['zf']
        q = q_args['q']
        q_line = lambda x: q*x/(q-1) - zf/(q-1)
        y_top = self.y_top
        x_bot = self.x_bot
        stages = Design['Theoretical stages']
        Rmin = Design['Minimum reflux']
        R = Design['Reflux']
        feed_stage = Design['Theoretical feed stage']
        
        # q_line
        intersect2 = lambda x: x - q_line(x)
        x_m2 = brentq(intersect2, 0, 1)
        
        # Graph q-line, Rectifying and Stripping section
        plt.plot([self._x_m, x_m2], [self._y_m, x_m2])
        plt.plot([self._x_m, y_top], [self._y_m, y_top])
        plt.plot([x_bot, self._x_m], [x_bot, self._y_m])
        plt.legend([f'Stages: {stages}, Feed: {feed_stage}', 'Graphical aid', 'eq-line', 'q-line', 'ROL', 'SOL'], fontsize=12)
        plt.title(f'McCabe Thiele Diagram (Rmin = {Rmin:.2f}, R = {R:.2f})')
        plt.show()
        return plt



# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:07:01 2018

@author: yoelr
"""
from ._tank import MixTank
from ._hx import HXutility

class EnzymeTreatment(MixTank):
    """Create an EnzymeTreatment unit that is cost as a MixTank with a heat exchanger."""
    _N_outs = 1
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None, *,
                 vessel_type='Conventional', tau=1.0, V_wf=0.9,
                 vessel_material='Stainless steel', T):
        MixTank.__init__(self, ID, ins, outs, thermo, vessel_type=vessel_type,
                         tau=tau, V_wf=V_wf, vessel_material=vessel_material)
        self.T = T #: Operating temperature
        self.heat_exchanger = hx = HXutility(None, None, T=T) 
        self.heat_utilities = hx.heat_utilities
        hx._ins = self._ins
        hx._outs = self._outs
    
    def _run(self):
        feed = self.ins[0]
        out = self.outs[0]
        out.mol[:] = self.mol_in
        out.phase = feed.phase
        out.P = feed.P
        out.T = self.T
        
    def _design(self):
        super()._design()
        self.heat_exchanger._design()
        
    def _cost(self):
        super()._cost()
        hx = self.heat_exchanger
        hx._cost()
        self.purchase_costs['Heat exchanger'] = hx.purchase_costs['Heat exchanger'] 
    


# -*- coding: utf-8 -*-
"""
Created on Sat Jul 13 02:24:35 2019

@author: yoelr
"""
from ._unit import Unit

__all__ = ('Facility',)

class Facility(Unit, isabstract=True, new_graphics=False):
    
    @property
    def system(self):
        return self._system
        


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:18:36 2018

@author: yoelr
"""
import numpy as np
from ..utils.design_warning import lb_warning
from .decorators import cost
from ._splitter import Splitter

@cost('Solids loading', cost=68040, CE=567, n=0.50, ub=40, BM=2.03,
      N='Number of centrifuges')
class SolidsCentrifuge(Splitter):
    """Create a solids centrifuge that separates out solids according to user defined split. Assume a continuous scroll solid bowl. 
    
    Parameters
    ----------
    ins : stream
        Inlet fluid with solids.
    outs : stream sequence
        * [0] Liquid-rich stream.
        * [1] Solids-rich stream.
    split: array_like
           Component splits to 0th output stream
    order=None : Iterable[str]
        Species order of split. Defaults to Stream.species.IDs.
    solids : tuple[str]
             IDs of solids.
    
    References
    ----------
        .. [0] Seider, Warren D., et al. (2017). "Cost Accounting and Capital Cost Estimation". In Product and Process Design Principles: Synthesis, Analysis, and Evaluation (pp. 481-485). New York: Wiley.
    
    """
    _units = {'Solids loading': 'tonn/hr'}
    _minimum_solids_loading = 2

    def __init__(self, ID='', ins=None, outs=(), *,
                 split, order=None, solids=None):
        super().__init__(ID, ins, outs, split=split, order=order)
        self.solids = solids
    
    @property
    def solids(self):
        return self._solids
    @solids.setter
    def solids(self, solids):
        self._solids = tuple(solids)
    
    def _design(self):
        mass_solids = 0
        solids = self._solids
        for s in self.ins:
            mass_solids += s.imass[solids]
        ts = np.asarray(mass_solids).sum() # Total solids
        ts *= 0.0011023 # To short tons (2000 lbs/hr)
        self.design_results['Solids loading'] = ts
        lb = self._minimum_solids_loading
        if ts < lb:
            lb_warning('Solids loading', ts, 'tonn/hr', lb)
    



# -*- coding: utf-8 -*-
"""
Created on Sat Nov 17 09:48:34 2018

@author: yoelr
"""
import numpy as np
import pandas as pd
from warnings import warn
import openpyxl
from .._tea import TEA, CombinedTEA
from thermosteam import Stream
from thermosteam.base import get_dimensionality, convert, stream_units_of_measure
from thermosteam.exceptions import DimensionError
import os

DataFrame = pd.DataFrame
ExcelWriter = pd.ExcelWriter

__all__ = ('stream_table', 'cost_table', 'save_system_results',
           'save_report', 'results_table', 'heat_utilities_table',
           'power_utilities_table')

def _stream_key(s):
    num = s.ID[1:]
    if num.isnumeric(): return int(num)
    else: return -1

# %% Helpful functions

def _save(tables, writer, sheet_name='Sheet1', n_row=1):
    """Save a list of tables as an excel file.
    
    Parameters
    ----------
    tables : iterable[DataFrame]
    writer : ExcelWritter
        
    """
    for t in tables:
        label = t.columns.name
        t.to_excel(writer, sheet_name, 
                   startrow=n_row, index_label=label)
        n_row += len(t.index) + 2
    
    return n_row

# %% Units

def save_report(system, file='report.xlsx', dpi='300', **stream_properties):
    """Save a system report as an xlsx file.
    
    Parameters
    ----------
    file : str
        File name to save report
    
    **stream_properties : str
        Additional stream properties and units as key-value pairs (e.g. T='degC', flow='gpm', H='kW', etc..)
        
    """
    writer = ExcelWriter(file)
    units = list(system._costunits)
    try:
        system.diagram('thorough', file='flowsheet', dpi=str(dpi), format='png')
    except:
        diagram_completed = False
        warn(RuntimeWarning('failed to generate diagram through graphviz'), stacklevel=2)
    else:
        try:
            # Assume openpyxl is used
            worksheet = writer.book.create_sheet('Flowsheet')
            flowsheet = openpyxl.drawing.image.Image('flowsheet.png')
            worksheet.add_image(flowsheet, anchor='A1')
        except:
            # Assume xlsx writer is used
            worksheet = writer.book.add_worksheet('Flowsheet')
            flowsheet.insert_image('A1', 'flowsheet.png')
        diagram_completed = True
    
    if system._TEA:
        tea = system._TEA
        if isinstance(tea, CombinedTEA):
            costs = [cost_table(i) for i in tea.TEAs]
            _save(costs, writer, 'Itemized costs')
        else:
            # Cost table
            cost = cost_table(tea)
            cost.to_excel(writer, 'Itemized costs')
        
        # Cash flow
        tea.get_cashflow_table().to_excel(writer, 'Cash flow')
    else:
        warn(RuntimeWarning(f'Cannot find TEA object in {repr(system)}. Ignoring TEA sheets.'), stacklevel=2)
    
    
    # Stream tables
    # Organize streams by chemicals first
    streams_by_chemicals = {}
    for i in system.streams:
        chemicals = i.chemicals
        if chemicals in streams_by_chemicals:
            streams_by_chemicals[chemicals].append(i)
        else:
            streams_by_chemicals[chemicals] = [i]
    stream_tables = []
    for streams in streams_by_chemicals.values():
        stream_tables.append(stream_table(streams, **stream_properties))
    _save(stream_tables, writer, 'Stream table')
    
    # Heat utility tables
    heat_utilities = heat_utilities_table(units)
    n_row = _save(heat_utilities, writer, 'Utilities')
    
    # Power utility table
    power_utility = power_utilities_table(units)
    power_utility.to_excel(writer, 'Utilities', 
                           index_label='Electricity',
                           startrow=n_row)
    
    # General desing requirements
    results = results_table(units)
    _save(results, writer, 'Design requirements')
    writer.save()
    if diagram_completed: os.remove("flowsheet.png")

save_system_results = save_report

def results_table(units):
    """Return a list of results tables for each unit type.

    Parameters
    ----------
    units : iterable[Unit]
        
    Returns
    -------
    tables : list[DataFrame]
    
    """
    units.sort(key=(lambda u: u.line))
    
    # Organize units by units of measure:
    organized = {}
    for u in units:
        uom = (*u._units.keys(), u.line)
        if uom in organized: organized[uom].append(u)
        else: organized[uom] = [u]
    
    # Make a list of tables, keeping all results with same keys in one table
    tables = []
    for units in organized.values():
        # First table with units of measure
        table = None
        while (table is None) and units:
            u, *units = units
            table = u.results(include_utilities=False,
                              include_total_cost=False)
        if table is None: continue
        for u in units[1:]:
            table[u.ID] = u.results(with_units=False, include_utilities=False,
                                   include_total_cost=False)
        table.columns.name = (u.line, '')
        tables.append(table)
        table = u.results()
    return tables
    
def cost_table(tea):
    """Return a cost table as a pandas DataFrame object.

    Parameters
    ----------
    units : iterable[Unit]
        
    Returns
    -------
    table : DataFrame

    """
    columns = ('Unit operation',
              f'Purchase cost (10^6 USD)',
              f'Utility cost (10^6 USD/yr)')
    units = tea.units
    operating_days = tea.operating_days
    N_units = len(units)
    array = np.empty((N_units, 3), dtype=object)
    IDs = []
    types = array[0:, 0]
    C_cap = array[0:, 1]
    C_op = array[0:, 2]
    
    # Get data
    for i in range(N_units):
        unit = units[i]
        types[i] = unit.line
        C_cap[i] = unit.purchase_cost / 1e6
        C_op[i] = unit.utility_cost * operating_days * 24  / 1e6
        IDs.append(unit.ID)
    
    df = DataFrame(array, columns=columns, index=IDs)    
    if not tea.lang_factor:
        df['Installation cost (10^6 USD)'] = [u.installation_cost for u in units]
    
    return df

def heat_utilities_table(units):
    """Return a list of utility tables for each heat utility source.
    
    Parameters
    ----------
    units : iterable[Unit]
        
    Returns
    -------
    tables : list[DataFrame]
        
    """
    # Sort heat utilities by unit type, then by utility Type
    units = sorted(units, key=(lambda u: type(u).__name__))
    source = {}
    heat_utils = []
    for u in units:
        hus = u.heat_utilities
        if not hus: continue
        for i in hus: source[i] = u
        heat_utils.extend(hus)

    # Organize heatutility by ID
    heat_utils_dict = {}
    for i in heat_utils:
        ID = i.ID
        if ID in heat_utils_dict: heat_utils_dict[ID].append(i)
        else: heat_utils_dict[ID] = [i]
    
    # First table and set Type to compare with
    hu = heat_utils[0]
    Type = hu.ID
    
    # Make a list of tables, keeping all results with same Type in one table
    tables = []
    for Type, heat_utils in heat_utils_dict.items():
        data = []; index = []
        for hu in heat_utils:
            data.append((source[hu].line, hu.duty, hu.flow, hu.cost))
            index.append(source[hu].ID)
        table = DataFrame(data, index=index,
                          columns=('Unit operation',
                                   'Duty (kJ/hr)',
                                   'Flow (kmol/hr)',
                                   'Cost (USD/hr)'))
        table.columns.name = Type
        tables.append(table)
    return tables
    

def power_utilities_table(units):
    # Sort power utilities by unit type
    units = sorted(units, key=(lambda u: type(u).__name__))
    units = [u for u in units if u.power_utility]
    power_utilities = [u.power_utility for u in units]
    lenght = len(power_utilities)
    data = []
    for i, u, pu in zip(range(lenght), units, power_utilities):
        data.append((u.line, pu.rate, pu.cost))
    return DataFrame(data, index=[u.ID for u in units if u.power_utility],
                     columns=('Unit Operation', 'Rate (kW)', 'Cost (USD/hr)'))


# %% Streams

def stream_table(streams, flow='kg/hr', **props) -> 'DataFrame':
    """Return a stream table as a pandas DataFrame object.

    Parameters

    streams : array_like[Stream]
    flow : str
        Units for flow rate.
    props : str
        Additional stream properties and units as key-value pairs
    
    """
    
    # Prepare rows and columns
    ss = sorted([i for i in streams if i.ID], key=_stream_key)
    chemical_IDs = ss[0].chemicals.IDs
    n = len(ss)
    m = len(chemical_IDs)
    p = len(props)
    array = np.empty((m+p+5, n), dtype=object)
    IDs = n*[None]
    sources = array[0, :]
    sinks = array[1, :]
    phases = array[2, :]
    prop_molar_data = array[3:3+p+1,:]
    flows = array[p+3, :]
    array[p+4, :] = ''
    fracs = array[p+5:m+p+5, :]
    for j in range(n):
        s = ss[j]
        sources[j] = str(s.source or '-')
        sinks[j] = str(s.sink or '-')
        IDs[j] = s.ID
        phase = ''
        for i in s.phase:
            if i == 'l':
                phase += 'liquid|'
            elif i == 'L':
                phase += 'LIQUID|'
            elif i == 'g':
                phase += 'gas|'
            elif i == 's':
                phase += 'solid|'
        phase = phase.rstrip('|')
        phases[j] = phase
        flow_j = s.get_flow(flow)
        flows[j] = net_j = sum(flow_j)
        fracs[:,j] = flow_j/net_j if net_j > 0 else 0
        i = 0
        for attr, units in props.items():
            prop_molar_data[i, j] = s.get_property(attr, units)
            i += 1
    
    # Set the right units
    i = 0
    prop_molar_keys = [f'{attr} ({unit})' for attr, unit in props.items()]
    
    # Make data frame object
    index = ('Source', 'Sink', 'Phase')  + tuple(prop_molar_keys) + (f'flow ({flow})', 'Composition:') + tuple(chemical_IDs)
    return DataFrame(array, columns=IDs, index=index)





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Apr 15 20:39:46 2018

@author: Yoel Rene Cortes-Pena
"""
from . import _flash
from . import _liquids_centrifuge
from .._graphics import Graphics
from ._mixer import Mixer
from ._splitter import Splitter, InvSplitter
from ._pump import Pump
from ._hx import HXutility, HXprocess
from ._flash import *
from ._flash import Flash
from ._multi_effect_evaporator import MultiEffectEvaporator
from ._liquids_centrifuge import *
from ._distillation import Distillation
from ._tank import Tank, StorageTank, MixTank
from ._transesterification import Transesterification
from ._fermentation import Fermentation
from ._enzyme_treatment import EnzymeTreatment
from ._clarifier import Clarifier
from ._solids_centrifuge import SolidsCentrifuge
from ._crushing_mill import CrushingMill
from ._rvf import RVF
from ._molecular_sieve import MolecularSieve
from ._balance import MassBalance
from ._conveying_belt import ConveyingBelt
from ._shredder import Shredder
from ._magnetic_separator import MagneticSeparator
from ._screw_feeder import ScrewFeeder
from ._vibrating_screen import VibratingScreen
from ._junction import Junction
from ._solids_separator import SolidsSeparator
from ._vent_scrubber import VentScrubber

# %% All units

__all__ = ['Mixer', 'Splitter', 'InvSplitter', 'MixTank', 'StorageTank', 'HXutility', 'HXprocess', 'Pump', 'Distillation', 'Transesterification', 'Fermentation', 'MultiEffectEvaporator', 'EnzymeTreatment', 'CrushingMill', 'RVF', 'MolecularSieve', 'SolidsCentrifuge', 'Clarifier', 'MassBalance', 'ConveyingBelt', 'Shredder', 'MagneticSeparator', 'ScrewFeeder', 'VibratingScreen', 'Junction', 'SolidsSeparator', 'VentScrubber']
__all__.extend(_flash.__all__)
__all__.extend(_liquids_centrifuge.__all__)

from lazypkg import LazyPkg
LazyPkg(__name__, ['decorators', 'design_tools', 'facilities'])


# %% Enhance Graphics

# Mixer
_mixgraphics = Mixer._graphics
_mixgraphics.edge_in = _mixgraphics.edge_in * 3
_mixgraphics.edge_out = _mixgraphics.edge_out * 3

# MixTank
_mixgraphics = MixTank._graphics
_mixgraphics.edge_in = _mixgraphics.edge_in * 3
_mixgraphics.edge_out = _mixgraphics.edge_out * 3

# Flash
edge_out = Flash._graphics.edge_out
edge_out[0]['tailport'] = 'n'
edge_out[1]['tailport'] = 's'
node = Flash._graphics.node
node['width'] = '1'
node['height'] = '1.1'

# Distillation
edge_out = Distillation._graphics.edge_out
edge_out[0]['tailport'] = 'n'
edge_out[1]['tailport'] = 's'
node = Distillation._graphics.node
node['width'] = '1'
node['height'] = '1.2'

# Single stream heat exchanger
HXutility._graphics = graphics = Graphics()
graphics.node['shape'] = 'circle'
graphics.node['color'] = 'none'
graphics.node['margin'] = '0'

def HXutility_node(hx):
    try:
        si = hx.ins[0]
        so = hx.outs[0]
        gi = si.phase == 'g'
        li = so.phase == 'l'
        go = so.phase == 'g'
        lo = so.phase == 'l'
        Ti = si.T
        To = so.T
        graphics = hx._graphics
        if Ti > To or (gi and lo):
            graphics.node['fillcolor'] = '#cfecf0'
            name = 'Cooling'
        elif Ti < To or (li and go):
            graphics.node['fillcolor'] = '#fad6d8'
            name = 'Heating'
        else:
            graphics.node['fillcolor'] = '#cfecf0:#fad6d8'
            name = 'Heat exchanger'
    except:
        graphics = hx._graphics
        name = 'Heat exchange'
    return name

graphics.node_function = HXutility_node

# Double stream heat exchanger
HXprocess._graphics = graphics = Graphics()
graphics.name = 'HXprocess'
graphics.node['shape'] = 'circle'
graphics.node['color'] = 'none'
graphics.node['margin'] = '0'
graphics.node['gradientangle'] = '90'
graphics.node['fillcolor'] = '#cfecf0:#fad6d8'

# Mixer
Mixer._graphics.node['shape'] = 'triangle'
Mixer._graphics.node['orientation'] = '270'
Mixer._graphics.edge_out[0]['tailport'] = 'e'

# Splitter
Splitter._graphics.node['shape'] = 'triangle'
Splitter._graphics.node['orientation'] = '90'
Splitter._graphics.node['fillcolor'] = "#bfbfbf:white"
Splitter._graphics.edge_in[0]['headport'] = 'w'

# Balance
graphics = MassBalance._graphics
graphics.node['shape'] = 'note'
graphics.node['fillcolor'] = '#F0F0F0'
graphics.in_system = False

del HXutility_node, graphics, edge_out, node, Graphics


# -*- coding: utf-8 -*-
"""
Created on Tue Mar 19 20:12:20 2019

@author: yoelr
"""

__all__ = []

from . import table
from . import plot

from .table import *
from .plot import *

__all__.extend(table.__all__)
__all__.extend(plot.__all__)


# -*- coding: utf-8 -*-
"""
Created on Sat Sep  1 17:35:28 2018

@author: yoelr
"""
import numpy as np
from ..utils.piping import StreamSequence
from .. import Unit, PowerUtility

__all__ = ('MassBalance',)

# %% Mass Balance Unit

class MassBalance(Unit):
    """Create a Unit object that changes net input flow rates to satisfy output flow rates. This calculation is based on mass balance equations for specified IDs. 

    Parameters
    ----------
    ins :
        Inlet streams.
    outs : 
        Outlet streams
    chemical_IDs : tuple[str]
        Chemicals that will be used to solve mass balance linear equations.
        The number of chemicals must be same as the number of input streams varied.
    streams : tuple[int]
        Indices of input streams that can vary in net flow rate
    is_exact=True : bool, optional
        True if exact flow rate solution is required for the specified IDs.
    balance='flow' : {'flow', 'composition'}, optional
          * 'flow': Satisfy output flow rates
          * 'composition': Satisfy net output molar composition

    Examples
    --------
    MassBalance are Unit objects that serve to alter flow rates of selected species and input streams to satisfy the mass balance. The example below uses the MassBalance object to satisfy the target flow rate feeding the mixer M1:
    
    >>> from biosteam import System
    >>> from biosteam.units import Mixer, Splitter, StorageTank, Pump, Flash, MassBalance
    >>> from thermosteam import Chemicals, Stream, settings
    >>> chemicals = Chemicals(['Water', 'Ethanol'])
    >>> settings.set_thermo(chemicals)
    >>> water = Stream('water',
    ...                Water=40,
    ...                units='lb/s',
    ...                T=350, P=101325)
    >>> ethanol = Stream('ethanol',
    ...                  Ethanol=190, Water=30,
    ...                  T=300, P=101325)
    >>> target = Stream('target', flow=[50, 50])
    >>> T1 = StorageTank('T1')
    >>> T2 = StorageTank('T2')
    >>> P1 = Pump('P1', P=101325)
    >>> P2 = Pump('P2', P=101325)
    >>> M1 = Mixer('M1', outs='s1')
    >>> S1 = Splitter('S1', outs=('s2', 's3'), split=0.5)
    >>> F1 = Flash('F1', outs=('s4', 's5'), V=0.5, P =101325)
    >>> # Connect units
    >>> water-T1-P1
    <Pump: P1>
    >>> ethanol-T2-P2
    <Pump: P2>
    >>> [P1-0, P2-0, S1-0]-M1-F1-1-S1
    <Splitter: S1>
    >>> MB1 = MassBalance('MB1', streams=[0,1],
    ...                   chemical_IDs=['Ethanol', 'Water'],
    ...                   outs=target,
    ...                   ins=(water, ethanol, S1-0))
    >>> mixSys = System('mixSys',
    ...                  recycle=S1-0,
    ...                  network=(MB1, T1, P1, T2, P2, M1, F1, S1))
    >>> # Make diagram to view system
    >>> # mixSys.diagram()
    >>> mixSys.simulate();
    >>> MB1.show()
    MassBalance: MB1
    ins...
    [0] water
        phase: 'l', T: 350 K, P: 101325 Pa
        flow (kmol/hr): Water  28.3
    [1] ethanol
        phase: 'l', T: 300 K, P: 101325 Pa
        flow (kmol/hr): Water    6.37
                        Ethanol  40.3
    [2] s2  from  Splitter-S1
        phase: 'l', T: 353.88 K, P: 101325 Pa
        flow (kmol/hr): Water    15.3
                        Ethanol  9.65
    outs...
    [0] target
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Water    50
                        Ethanol  50
    
    """
    line = 'Balance'
    results = None
    _has_cost = False
    _N_ins = 2
    heat_utilities = ()
    power_utility = None

    def __init__(self, ID='', outs=(), ins=None,
                 chemical_IDs=None, streams=None,
                 is_exact=True, balance='flow', thermo=None):
        self.ID = ID
        thermo = self._load_thermo(thermo)
        self._ins = StreamSequence(self._N_ins, ins, thermo, fixed_size=False)
        self._outs = StreamSequence(self._N_outs, outs, thermo, fixed_size=False)
        self._power_utility = PowerUtility()
        self.streams = streams
        self.chemical_IDs = chemical_IDs
        self.is_exact = is_exact
        self.balance = balance
        
    def _run(self):
        """Solve mass balance by iteration."""
        # SOLVING BY ITERATION TAKES 15 LOOPS FOR 2 STREAMS
        # SOLVING BY LEAST-SQUARES TAKES 40 LOOPS
        stream_index = self.streams
        balance = self.balance
        solver = np.linalg.solve if self.is_exact else np.linalg.lstsq
        ins = self.ins

        # Set up constant and variable streams
        vary = [ins[i] for i in stream_index]  # Streams to vary in mass balance
        constant = [i for i in ins if (i and i not in vary)]  # Constant streams
        index = self.chemicals.get_index(self.chemical_IDs)

        if balance == 'flow':
            # Perform the following calculation: Ax = b = f - g
            # Where:
            #    A = flow rate array
            #    x = factors
            #    b = target flow rates
            #    f = output flow rates
            #    g = constant input flow rates

            # Solve linear equations for mass balance
            A = np.array([s.mol for s in vary]).transpose()[index, :]
            f = self.mol_out[index]
            g = sum([s.mol[index] for s in constant])
            b = f - g
            x = solver(A, b)

            # Set flow rates for input streams
            for factor, s in zip(x, vary):
                s.mol[:] = s.mol * factor

        elif balance == 'fraction':
            # Perform the following calculation:
            # Ax = b
            #    = sum( A_ * x_guess + g_ )f - g
            #    = A_ * x_guess * f - O
            # O  = sum(g_)*f - g
            # Where:
            # A_ is flow array for all species
            # g_ is constant flows for all species
            # Same variable definitions as in 'flow'

            # Set all variables
            A_ = np.array([s.mol for s in vary]).transpose()
            A = np.array([s.mol for s in vary]).transpose()[index, :]
            f = self.z_mol_out[index]
            g_ = sum([s.mol for s in constant])
            g = g_[index]
            O = sum(g_) * f - g

            # Solve by iteration
            x_guess = np.ones_like(index)
            not_converged = True
            while not_converged:
                # Solve linear equations for mass balance
                b = (A_ * x_guess).sum()*f + O
                x_new = solver(A, b)
                not_converged = sum(((x_new - x_guess)/x_new)**2) > 0.0001
                x_guess = x_new

            # Set flow rates for input streams
            for factor, s in zip(x_new, vary):
                s.mol = s.mol * factor
        
        else:
            raise ValueError( "balance type must be one of the following: 'flow', 'composition'")


# %% Energy Balance Unit

# class EnergyBalance(Unit):
#     """Create a Unit object that changes a stream's temperature, flow rate, or vapor fraction to satisfy energy balance.

#     **Parameters**

#         **index:** [int] Index of stream that can vary in temperature, flow rate, or vapor fraction.
        
#         **Type:** [str] Should be one of the following
#             * 'T': Vary temperature of output stream
#             * 'F': Vary flow rate of input/output stream
#             * 'V': Vary vapor fraction of output stream
        
#         **Qin:** *[float]* Additional energy input.
        
#     .. Note:: This is not a mixer, input streams and output streams should match flow rates.

#     """
#     _kwargs = {'index': None,
#                'Type': 'T',
#                'Qin': 0}
#     line = 'Balance'
#     _has_cost = False
#     _graphics = MassBalance._graphics
#     _init_ins = MassBalance._init_ins
#     _init_outs = MassBalance._init_outs
    
#     def _run(self):        # Get arguments
#         ins = self.ins.copy()
#         outs = self.outs.copy()
#         kwargs = self._kwargs
#         index = kwargs['index']
#         Type = kwargs['Type']
#         Qin = kwargs['Qin']
        
#         # Pop out required streams
#         if Type == 'F':
#             s_in = ins.pop(index)
#             s_out = outs.pop(index)
#         else:
#             s = outs.pop(index)
        
#         # Find required enthalpy
#         H_in = sum(i.H for i in ins) + Qin
#         H_out = sum(o.H for o in outs)
#         H_s = H_out - H_in
        
#         # Set enthalpy
#         if Type == 'T':
#             s.H = -H_s
#         elif Type == 'V':
#             s.enable_phases()
#             s.VLE(Qin=s.H - H_s)
#         elif Type == 'F':
#             s.mol *= (s_out.H - s_in.H)/H_s
#         else:
#             raise ValueError(f"Type must be 'T', 'V' or 'F', not '{Type}'")
            
        
        


# -*- coding: utf-8 -*-
"""
Created on Tue Feb 11 02:11:08 2020

@author: yoelr
"""
__all__ = ('ExponentialFunctor',)

class ExponentialFunctor:
    """
    Create an ExponentialFunctor object for computing equations of the 
    form :math:`f(S) = A \cdot S^n`.
    
    Parameters
    ----------
    A : float
        Linear coefficient.
    n : float
        Exponential coefficient.
    
    Attributes
    ----------
    A : float
        Linear coefficient.
    n : float
        Exponential coefficient.
        
    Examples
    --------
    >>> from biosteam.utils import ExponentialFunctor
    >>> f_exp = ExponentialFunctor(A=10.0, n=2.0)
    >>> f_exp(2.0)
    40.0
    """    
    __slots__ = ('A', 'n')

    def __init__(self, A, n):
        self.A = A
        self.n = n

    def __call__(self, S):
        return self.A * S ** self.n

    def __repr__(self):
        return f"{type(self).__name__}(A={self.A}, n={self.n})"


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 21:43:13 2018

@author: yoelr
"""
import numpy as np
import biosteam as bst
from .. import Unit
from . import Mixer, HXutility
from ._flash import Evaporator_PV, Evaporator_PQ
from .design_tools import (
    compute_vacuum_system_power_and_cost,
    compute_heat_transfer_area
)    
from thermosteam import Stream, settings
from flexsolve import IQ_interpolation
from warnings import warn
import ht
log = np.log
exp = np.exp

# Table 22.32 Product process and design (pg 592)
# Name: ('Area range (m2)', 'Cost(A) (USD)', 'U (kJ/(hr*m2*K)))', 'Material')
evaporators = {'Horizontal tube': 
                    ((9.29, 743.224),
                     lambda A, CE: CE*2.304*A**0.53,
                     4906.02,
                     'Carbon steel'),
               'Long-tube vertical':
                   ((9.29, 743.224),
                    lambda A, CE: CE*3.086*A**0.55,
                    8176.699,
                    'Carbon steel'),
               'Forced circulation': 
                   ((13.935, 8000),
                    lambda A, CE: CE/500*exp(8.2986 + 0.5329*log(A*0.0929)-0.000196*log(A*0.0929)**2),
                    10731.918,
                    'Carbon steel'),
               'Falling film': 
                   ((13.935, 371.612),
                    lambda A, CE: CE*7.416*A**0.55,
                    10220.874,
                    'Stainless steel tubes/Carbon steel shell')}


class MultiEffectEvaporator(Unit):
    """Creates evaporatorators with pressures given by P (a list of pressures). Adjusts first evaporator vapor fraction to satisfy an overall fraction evaporated. All evaporators after the first have zero duty. Condenses the vapor coming out of the last evaporator. Pumps all liquid streams to prevent back flow in later parts. All liquid evaporated is ultimately recondensed. Cost is based on required heat transfer area. Vacuum system is based on air leakage. Air leakage is based on volume, as given by residence time `tau` and flow rate to each evaporator.

    Parameters
    ----------
    ins : stream
        Inlet.
    outs : stream sequence
        * [0] Solid-rich stream.
        * [1] Condensate stream.
    component : str
                Component being evaporated.
    P : tuple[float]
        Pressures describing each evaporator (Pa).
    V : float
        Overall molar fraction of component evaporated.
    P_liq : tuple
            Liquid pressure after pumping (Pa).
    
    """
    _units = {'Area': 'm^2',
              'Volume': 'm^3'}
    _N_outs = 2
    _N_heat_utilities = 2
    BM = 2.45
    line = 'Multi-Effect Evaporator'

    #: Residence time (hr)
    tau = 0.30

    # Evaporator type
    _Type = 'Forced circulation'
    
    # Data for simmulation and costing
    _evap_data = evaporators[_Type]

    @property
    def Type(self):
        """Evaporation type."""
        return self._Type
    @Type.setter
    def Type(self, evap_type):
        try:
            self._evap_data = evaporators[evap_type]
        except KeyError:
            dummy = str(evaporators.keys())[11:-2]
            raise ValueError(f"Type must be one of the following: {dummy}")
        self._Type = evap_type

    def __init__(self, ID='', ins=None, outs=(), *, P, V):
        Unit.__init__(self, ID, ins, outs)
        # Unpack
        out_wt_solids, liq = self.outs
        self.V = V #: [float] Overall molar fraction of component evaporated.
        self._V1 = V/2.
        
        # Create components
        self._N_evap = n = len(P) # Number of evaporators
        first_evaporator = Evaporator_PV(None, outs=(None, None), P=P[0])
        
        # Put liquid first, then vapor side stream
        evaporators = [first_evaporator]
        for i in range(1, n):
            evap = Evaporator_PQ(None, outs=(None, None, None), P=P[i], Q=0)
            evaporators.append(evap)
        
        condenser = HXutility(None, outs=Stream(None), V=0)
        self.heat_utilities = (first_evaporator.heat_utilities[0],
                               condenser.heat_utilities[0])
        mixer = Mixer(None, outs=Stream(None))
        
        self.components = {'evaporators': evaporators,
                           'condenser': condenser,
                           'mixer': mixer}
        

    def _run(self):
        out_wt_solids, liq = self.outs
        ins = self.ins

        n = self._N_evap  # Number of evaporators

        # Set-up components
        components = self.components
        evaporators = components['evaporators']
        first_evaporator, *other_evaporators = evaporators
        first_evaporator.ins[:] = [i.copy() for i in ins]
        condenser = components['condenser']
        mixer = components['mixer']
        
        # Put liquid first, then vapor side stream
        ins = [first_evaporator.outs[1], first_evaporator.outs[0]]
        for evap in other_evaporators:
            evap.ins[:] = ins
            ins = [evap.outs[1], evap.outs[0]]
        
        def compute_overall_vapor_fraction(v1):
            v_overall = v1
            first_evaporator.V = v1
            first_evaporator._run()
            for evap in other_evaporators:
                evap._run()
                v_overall += (1-v_overall) * evap.V
            return v_overall
        
        x0 = 0.0001
        x1 = 0.9990
        y0 = compute_overall_vapor_fraction(x0)
        y1 = compute_overall_vapor_fraction(x1)
        self._V1 = IQ_interpolation(compute_overall_vapor_fraction,
                                    x0, x1, y0, y1, self._V1, self.V, 
                                    xtol=0.0001, ytol=0.001)
        # Condensing vapor from last effector
        outs_vap = evaporators[-1].outs[0]
        condenser.ins[:] = [outs_vap]
        condenser._run()
        outs_liq = [condenser.outs[0]]  # list containing all output liquids

        # Unpack other output streams
        out_wt_solids.copy_like(evaporators[-1].outs[1])
        for i in range(1, n):
            evap = evaporators[i]
            outs_liq.append(evap.outs[2])

        # Mix liquid streams
        mixer.ins[:] = outs_liq
        mixer._run()
        liq.copy_like(mixer.outs[0])
        
    def _design(self):
        # This functions also finds the cost
        A_range, C_func, U, _ = self._evap_data
        components = self.components
        evaporators = components['evaporators']
        Design = self.design_results
        Cost = self.purchase_costs
        CE = bst.CE
        
        first_evaporator = evaporators[0]
        hu = first_evaporator.heat_utilities[0]
        duty = first_evaporator.H_out - first_evaporator.H_in
        Q = abs(duty)
        Tci = first_evaporator.ins[0].T
        Tco = first_evaporator.outs[0].T
        hu(duty, Tci, Tco)
        Th = hu.inlet_utility_stream.T
        LMTD = ht.LMTD(Th, Th, Tci, Tco)
        ft = 1
        A = abs(compute_heat_transfer_area(LMTD, U, Q, ft))
        self._evap_costs = evap_costs = [C_func(A, CE)]
        
        # Find condenser requirements
        condenser = components['condenser']
        condenser._design()
        condenser._cost()
        Cost['Condenser'] = condenser.purchase_costs['Heat exchanger']
        
        # Find area and cost of evaporators
        As = [A]
        A_min, A_max = A_range
        for evap in evaporators[1:]:
            Q = evap._Q
            Tc = evap.outs[0].T
            Th = evap.outs[2].T
            LMTD = Th - Tc
            A = compute_heat_transfer_area(LMTD, U, Q, ft)
            As.append(A)
            if settings.debug and not A_min < A < A_max:
                warn(f'area requirement ({A}) is out of range, {A_range}')
            evap_costs.append(C_func(A, CE))
        self._As = As
        Design['Area'] = A = sum(As)
        Design['Volume'] = total_volume = self._N_evap * self.tau * self.ins[0].F_vol
        Cost['Evaporators'] = sum(evap_costs)
        
        # Calculate power
        power, cost = compute_vacuum_system_power_and_cost(
            F_mass=0, F_vol=0, P_suction=evap.outs[0].P,
            vessel_volume=total_volume,
            vacuum_system_preference='Liquid-ring pump')
        Cost['Vacuum liquid-ring pump'] = cost
        self.power_utility(power)
        
        



# -*- coding: utf-8 -*-
"""
Created on Sun Dec  8 02:13:07 2019

@author: yoelr
"""
from biosteam import units
import doctest

__all__ = ('test_binary_distillation', 'test_mixer', 'test_splitter',
           'test_tank')

def test_binary_distillation():
    doctest.testmod(units._distillation)
    
def test_mixer():
    doctest.testmod(units._mixer)

def test_splitter():
    doctest.testmod(units._splitter)
    
def test_tank():
    doctest.testmod(units._tank)
    
def test_fermentation():
    doctest.testmod(units._fermentation)

def test_heat_exchanger():
    doctest.testmod(units._hx)

def test_pump():
    doctest.testmod(units._pump)

def test_molecular_sieve():
    doctest.testmod(units._molecular_sieve)

def test_flash():
    doctest.testmod(units._flash)
    
def test_mass_balance():
    doctest.testmod(units._balance)
    
def test_unit_operations():
    test_binary_distillation()
    test_mixer()
    test_splitter()
    test_tank()
    test_fermentation()
    test_heat_exchanger()
    test_pump()
    test_molecular_sieve()
    test_flash()
    test_mass_balance()
    
if __name__ == '__main__':
    test_unit_operations()


import matplotlib.pyplot as plt
import matplotlib.patches as patches
from .biosteam_colors import colors
import numpy as np

__all__ = ('DoubleColorCircle', 'DoubleColorLegend')

# %% Legend handlers

def make_patch(pos, width, height, Patch, handlepatch, kwargs):
    if issubclass(Patch , patches.Circle): 
        patch = Patch(pos+[width/2,height/2], height/2,
                      transform=handlepatch.get_transform(), **kwargs)    
    elif issubclass(Patch, patches.Rectangle): 
        patch = Patch(pos, width, height,
                      transform=handlepatch.get_transform(), **kwargs)    
    else:
        raise NotImplementedError(f"patch '{Patch}' not implemented")
    handlepatch.add_artist(patch)
    return patch

class DoubleColor:
    __slots__ = ('leftpatches', 'rightpatches', 'Patch', 'patches', 'both')
    
    def __init__(self, leftpatches, rightpatches, morepatches=None,
                 shape='Rectangle', both={}):
        self.leftpatches = leftpatches
        self.rightpatches = rightpatches
        self.Patch = getattr(patches, shape)
        self.patches = morepatches or []
        self.both = both
        
    def legend_artist(self, legend, handle, fontsize, patch):
        x0, y0 = patch.xdescent, patch.ydescent
        width, height = patch.width/2, patch.height
        leftpos = np.array([x0, y0])
        rightpos = np.array([x0+width, y0])
        Patch = self.Patch
        for l in self.leftpatches:
            lpatch = make_patch(leftpos, width, height, Patch, patch, l)
        for r in self.rightpatches:
            rpatch = make_patch(rightpos, width, height, Patch, patch, r)
            patch.add_artist(rpatch)
        for patch in self.patches:
            patch.legend_artist(legend, handle, fontsize, patch)
        both = self.both
        if both:
            if 'fill' not in both:
                both['fill'] = False
            if 'edgecolor' not in both:
                both['edgecolor'] = 'k'
            patch.add_artist(patches.Rectangle(leftpos, width*2, height,
                transform=patch.get_transform(), **both))
        return lpatch


class DoubleColorCircle:
    __slots__ = ('left', 'right', 'both')
    
    def __init__(self, left, right, both):
        self.left = left
        self.right = right
        self.both = both
        
    def legend_artist(self, legend, handle, fontsize, patch):
        x0, y0 = patch.xdescent, patch.ydescent
        width, height = patch.width, patch.height
        pos = np.array([x0+width/2, y0+height/2])
        leftpatch = patches.Arc(pos, width/2, width/2,
                                angle=.0, theta1= 90.0, theta2=-90.0,
                                **self.left, hatch = 20*'o',
                                transform=patch.get_transform())
        rightpatch = patches.Arc(pos, width/2, width/2,
                                angle=.0, theta1=-90.0, theta2= 90.0,
                                **self.right, hatch = 20*'o',
                                transform=patch.get_transform())
        patch.add_artist(leftpatch)
        patch.add_artist(rightpatch)
        both = self.both
        if both:
            if 'fill' not in both:
                both['fill'] = False
            if 'edgecolor' not in both:
                both['edgecolor'] = 'k'
            patch.add_artist(patches.Arc(pos, width/2, width/2,
                                         transform=patch.get_transform(),
                                         **both))
        return leftpatch


class TwoColorArrow:
    __slots__ = ('left', 'right', 'pos')
    
    def __init__(self, left, right, pos='mid'):
        self.left = left
        self.right = right
        self.pos = pos
        
    def legend_artist(self, legend, handle, fontsize, patch):
        x0, y0 = patch.xdescent, patch.ydescent
        width = patch.width
        height = patch.height
        pos = self.pos
        if pos=='mid':
            y0 += height/2
        elif pos=='bot':
            pass
        elif pos=='top':
            y0 += height
        else:
            ValueError('invalid pos in ArrowPatch object')
        leftpatch = patches.Arrow(x0, y0, width/2, 0,
                                  **self.left)
        rightpatch = patches.Arrow(x0+width/2, y0, width/2, 0,
                                   **self.right)
        patch.add_artist(leftpatch)
        patch.add_artist(rightpatch)
        return leftpatch


class Legend:
    __slots__ = ('handler_map')
    
    def __init__(self, handler_map=None):
        self.handler_map = handler_map or {}
    
    def legend(self, loc='upper right', **kwargs):
        hmap = self.handler_map
        names = hmap.keys()
        plt.legend(names, names, handler_map=self.handler_map, prop=kwargs,
                   loc=loc)


class DoubleColorLegend(Legend):
    __slots__ = ()
    
    def __init__(self, key_leftrightpatches=None, handler_map=None):
        self.handler_map = handler_map or {}
        if key_leftrightpatches:
            for key, left_right in key_leftrightpatches.items():
                handler_map[key] = DoubleColor(*left_right)
    
    def add_box(self, name,
                leftcolor=colors.orange_tint.RGBn, 
                rightcolor=colors.blue_tint.RGBn,
                both={}):  
        self.handler_map[name] = DoubleColor([{'facecolor': leftcolor}],
                                             [{'facecolor': rightcolor}],
                                             None, 'Rectangle', both=both)

    def add_circle(self, name,
                   leftcolor=colors.orange_shade.RGBn,
                   rightcolor=colors.blue_shade.RGBn,
                   both={}):
        self.handler_map[name] = DoubleColorCircle(
                   {'color': leftcolor,
                    'edgecolor': 'black'},
                   {'color': rightcolor,
                    'edgecolor': 'black'},
                    both=both)
    
    # def add2key(self, key, patch, side='left'):
    #     db = self.handler_map[key]
    #     sidepatches = getattr(db, side+'patches')
    #     sidepatches.append(patch)
    
    # def newkey(self, key, leftpatches, rightpatches, patches=None, shape='Rectangle', **kwargs):
    #     if isinstance(leftpatches, dict):
    #         leftpatches = [leftpatches]
    #     if isinstance(rightpatches, dict):
    #         rightpatches = [rightpatches]
    #     self.handler_map[key] = DoubleColor(leftpatches, rightpatches, patches, shape, **kwargs)